#pragma once

enum class MViewAction
{
  Store = 0,
  Clear = 1,
  Interpolate = 2,
  Reinterpolate = 3,
  Smooth = 4,
  Reset = 5,
  Uninterpolate = 6,
  Toggle = 7,
  ToggleInterp = 8,
  Purge = 9
};
